require 'spec_helper'
require 'stub_user'
require 'fakeweb'
#require 'webmock/rspec'
require 'active_resource'
require 'active_resource/http_mock'

# admin@example.org
# 6a002a40-bc63-4095-9bb0-f31bf386bf55
# admin

# nobody@example.org
# 596568c5-d6e7-4328-815c-c532aeed1198
# no role
 
describe DisiidUser do
  
  before { @user = User.new }
    
  it 'should return correct version string' do
    DisiidUser.version_string.should == "DisiidUser version #{DisiidUser::VERSION}"
  end
  
  it 'should extract uuid from identity_url' do
    @user.identity_url = "http://openid.provider.local/users/6a002a40-bc63-4095-9bb0-f31bf386bf55"
    @user.uuid.should == "6a002a40-bc63-4095-9bb0-f31bf386bf55"
  end
  
  it 'should return nil if identity_url is of unknown nature' do
    @user.identity_url = "http://openid.provider.local/6a002a40-bc63-4095-9bb0-f31bf386bf55"
    @user.uuid.should be_nil
  end
  
  it 'should return nil if identity_url is not defined' do
    @user.identity_url = ""
    @user.uuid.should be_nil
    
    # uuid is not memorized so we're safe to reuse @user instance
    @user.identity_url = nil
    @user.uuid.should be_nil
  end
  
  shared_examples_for "new user from identity" do
    it { should be_instance_of(User) }
    it { should be_new_record }
#    its(:identity_url) { should == 'http://openid.provider.local/users/6a002a40-bc63-4095-9bb0-f31bf386bf55' }
#    its(:uuid) { should == '6a002a40-bc63-4095-9bb0-f31bf386bf55' }
    it "has correct identity url" do
      expect(subject.identity_url).to eq 'http://openid.provider.local/users/6a002a40-bc63-4095-9bb0-f31bf386bf55'
    end
    it "has correct uuid" do
      expect(subject.uuid).to eq '6a002a40-bc63-4095-9bb0-f31bf386bf55'
    end
  end
  
  describe ".build_from_identity" do
    it_should_behave_like "new user from identity" do
      subject { User.build_from_identity('http://openid.provider.local/users/6a002a40-bc63-4095-9bb0-f31bf386bf55') }
    end
  end
  
  describe ".build_from_identity_url" do
    it_should_behave_like "new user from identity" do
      subject { User.build_from_identity_url('http://openid.provider.local/users/6a002a40-bc63-4095-9bb0-f31bf386bf55') }
    end
  end
  
  describe "#find_local_or_remote_by_uuid!" do
    let(:user) { User }
    subject { user }
    
    it { should respond_to :find_local_or_remote_by_uuid! }
    it "should look for a local user" do
      expect(User).to receive(:where).and_return([@user])
      user.find_local_or_remote_by_uuid!('some-uuid')
    end
    
    context "when local user doesn't exist" do
      let(:new_user) { mock('new user', :save! => true) }
      before { 
        user.stub(:where).and_return([]) 
        user.stub(:build_from_identity).and_return(new_user)
      }
      
      it "should try and see if remote user exists" do
        DisiidUser::RemoteUser.should_receive(:exists?).with('some-uuid', params: hash_including(
          auth_token: DisiidUser::RemoteUser.auth_token
        ))
        user.find_local_or_remote_by_uuid!('some-uuid')
      end
      
      context "but the remote does" do
        before { DisiidUser::RemoteUser.stub(:exists?).and_return(true) }
        it "should build a new user" do
          user.should_receive(:build_from_identity).with('http://localhost:3001/users/a-uuid')
          user.find_local_or_remote_by_uuid!('a-uuid')
        end
        
        it "should store user with .save!" do
          new_user.should_receive(:save!)
          user.find_local_or_remote_by_uuid!('a-uuid')
        end
        
        it "should return newly created user" do
          user.find_local_or_remote_by_uuid!('a-uuid').should == new_user
        end
        
        it "should raise 'record not found' if .save! fails" do
          new_user.stub(:save!).and_raise('some exception')
          expect { user.find_local_or_remote_by_uuid!('some-uuid') }.to raise_error(ActiveRecord::RecordNotFound)
        end
      end
      
      context "and niether does remote" do
        before { DisiidUser::RemoteUser.stub(:exists?).and_return(false) }
        subject { user.find_local_or_remote_by_uuid!('some-uuid') }
        it { should be_nil }
      end
      
      context "and an exception's been raised" do
        before { DisiidUser::RemoteUser.stub(:exists?).and_raise('some error') }
        it "should throw 'record not found'" do
          expect { user.find_local_or_remote_by_uuid!('some-uuid') }.to raise_error(ActiveRecord::RecordNotFound)
        end
      end
    end
    
    context "when local user exists" do
      let(:existing_user) { mock('existing user') }
      before { user.stub(:where).and_return([existing_user]) }
      
      it "should not fetch from remote" do
        DisiidUser::RemoteUser.should_not_receive(:exists?)
        user.should_not_receive(:build_from_identity)
        user.find_local_or_remote_by_uuid!('some-uuid').should == existing_user
      end
    end
  end
  
  context "with local roles" do
    
    it 'should correctly parse :role attribute' do
      @user.role = 1
      @user.local_roles.should include 'admin'
      @user.local_roles.length.should == 1
      
      @user.role = 3
      @user.local_roles.should include 'admin'
      @user.local_roles.should include 'manager'
      @user.local_roles.length.should == 2
    end
    
    it 'should correctly set :role attribute' do
      @user.local_roles = 'admin', 'manager'
      @user.role.should == 3
      
      @user.local_roles = ['manager']
      @user.role.should == 2
    end
    
    describe 'has_role?() method' do
    
      it 'should use roles()' do
        @user.should_receive(:roles).at_least(:once).and_return([])
        @user.has_role?("does't matter")
      end
      
      context 'when admin' do
        before { @user.role = 1 }
        
        it 'should always return true' do
          @user.has_role?('admin').should be true
          @user.has_role?('manager').should be true
        end
      end
      
      context 'when manager' do
        before { @user.local_roles = :manager }
        
        it 'should not be admin' do
          @user.has_role?(:admin).should_not be true
        end
        
        it 'should not be something else' do
          @user.has_role?('something').should_not be true
        end
        
        it 'should be manager' do
          @user.has_role?('manager').should be true
        end
      end # context 'when manager'
    end # describe 'has_role?() method'
  end # context "with local roles"
  
  context 'with remote user' do
    before { DisiidUser::RemoteUser.auth_token = 'sometoken' }
    
    it 'should fetch it using find()' do
      @user.identity_url = "http://host/users/some-uuid"
      DisiidUser::RemoteUser.should_receive(:find).with('some-uuid', :params => hash_including(:auth_token => 'sometoken'))
      
      @user.remote_user
    end
    
    it 'should be cached' do
      @user.identity_url = "http://host/users/some-uuid"
      
      cache_expiry = DisiidUser::RemoteUser.cache_expiry
      Rails.cache.should_receive(:fetch).with('some-uuid', hash_including(:expires_in => cache_expiry))
      
      @user.remote_user
    end
    
    it 'should not search for it if uuid is missing' do
      @user.identity_url = nil
      DisiidUser::RemoteUser.should_not_receive(:find)
      
      @user.remote_user.should be_nil
    end
    
    context 'when does not exist' do
      before { @user.identity_url = "http://openid.provider.local/users/does-no-exist" }
      it 'should just be nil' do
        @user.remote_user.should be_nil
      end
      
      it 'should return nil for remote attributes' do
        @user.email.should be_nil
        @user.homepage.should be_nil
        @user.first_name.should be_nil
        @user.last_name.should be_nil
        @user.nickname.should be_nil
        @user.homepage.should be_nil
        @user.position.should be_nil
        @user.office_address.should be_nil
        @user.office_num.should be_nil
        @user.office_phone.should be_nil
        @user.office_fax.should be_nil
        @user.provider.should be_nil
        @user.ada_id.should be_nil
        @user.okkam_id.should be_nil
        @user.remote_created_at.should be_nil 
        @user.remote_updated_at.should be_nil
      end
      
      it 'should not raise an exception on attributes assignment' do
        @user.first_name = 'whatever'
        # this will basically check the part '... if remote_user'
        # in e.g. define_method('first_name=') { |val| first_name = val if remote_user }
        @user.first_name.should be_nil
        # assuming other attributes work the same way
      end
    end
    
    context 'when exists' do
      
      before {
        @user.identity_url = "http://openid.provider.local/users/6a002a40-bc63-4095-9bb0-f31bf386bf55"
        ActiveResource::HttpMock.respond_to do |mock|
             mock.get "/users/#{@user.uuid}.xml?auth_token=#{DisiidUser::RemoteUser.auth_token}", {"Accept" => "application/xml"}, File.read("#{File.dirname(__FILE__)}/#{@user.uuid}.xml"), 200
            mock.get "/users/#{@user.uuid}.json?auth_token=#{DisiidUser::RemoteUser.auth_token}", {"Accept" => "application/json"}, File.read("#{File.dirname(__FILE__)}/#{@user.uuid}.json"), 200
          end
      }
    
      [:xml, :json].each { |format|
        it "should have remote attributes in #{format} format" do
          DisiidUser::RemoteUser.format = format
          DisiidUser::RemoteUser.auth_token = 'sometoken'
          ActiveSupport.parse_json_times = true
          expect(@user.email).not_to be_nil
          @user.email.should == 'admin@example.org'
          @user.first_name.should == 'John'
          @user.last_name.should == 'Smith'
          @user.homepage.should == 'http://john.smith.example.org'
          @user.nickname.should == 'john.smith'
          @user.position.should == 'full_proffesor'
          @user.office_address.should == 'via sommarive, 5'
          @user.office_num.should == 'D14'
          @user.office_phone.should == '0461012345'
          @user.office_fax.should == '0461540123'
          @user.provider.should be_nil
          @user.unitn_uid.should be_nil
          @user.ada_id.should == 'ADA1122334455'
          @user.okkam_id.should == 'kkilpp009iiouy7rccfw5aa'
          @user.photo.should == 'http://remote_url.example.com/assets/nophoto.jpg'
          @user.photo_normal.should == 'http://remote_url.example.com/assets/nophoto-normal.jpg'
          @user.photo_thumb.should == 'http://remote_url.example.com/assets/nophoto-thumb.jpg'
          # in JSON the following two will be strings
          # if ActiveSupport.parse_json_times is not set to true
          @user.remote_created_at.should == DateTime.parse('2011-11-03T10:23:11Z')
          @user.remote_updated_at.should == DateTime.parse('2011-12-06T11:15:33Z')
          # for this to evaluate to boolean 
          # XML tag should contain type="boolean" attribute
          expect(@user.hide_photo).to be false
        end
      }
      
      it 'should have assign methods like email=(...)' do
        DisiidUser::RemoteUser.format = :xml
        DisiidUser::RemoteUser.auth_token = 'sometoken'
        DisiidUser::RemoteUser.site = "http://openid.provider.local"
        
        @user.email ='new@email.com'
        expect(@user.email).to eq 'new@email.com'

        @user.homepage = 'http://some.page.com'
        @user.homepage.should == 'http://some.page.com'
        # assuming other attributes work too
      end
    
      it 'should get all roles, including remote' do
        @user.local_roles = 'manager'
        
        @user.roles.should include 'admin' # remote
        @user.roles.should include 'manager' # local
      end
    
      it 'should not have duplicated roles' do
        @user.local_roles = 'admin', 'manager'
        # this remote user has two roles, :admin and :whatever
        # spec/6a002a40-bc63-4095-9bb0-f31bf386bf55.xml
        %w(admin manager whatever).each { |r|
          @user.roles.should include r
        }
        @user.roles.length.should == 3 
      end
    end # context 'when exists' 
    
    context 'when search' do
      before { 
        DisiidUser::RemoteUser.auth_token = 'sometoken'
        ActiveResource::HttpMock.respond_to do |mock|
             mock.get "/users.xml?auth_token=#{DisiidUser::RemoteUser.auth_token}&q=e", {"Accept" => "application/xml"}, File.read("#{File.dirname(__FILE__)}/users.xml"), 200
            mock.get "/users.json?auth_token=#{DisiidUser::RemoteUser.auth_token}&q=e", {"Accept" => "application/json"}, File.read("#{File.dirname(__FILE__)}/users.json"), 200
        end
      
      }
    
      it 'should find a list of users with activeresource' do
        @users = DisiidUser::RemoteUser.find(:all , :params => {:q => 'e', :auth_token => 'sometoken'})
        @users.size.should == 3
        @users[0].uuid.should == "7ee5003f-faaa-40ac-939c-615c0da69deb"
        @users[1].uuid.should == "2632c516-eb32-4cc8-879f-e8060577fc14"
        @users[2].uuid.should == "ad80daad-7a2f-4f5e-a62e-3189248486a9"
        @users[0].photo.should == "http://localhost:3000/uploads/user/photo/7ee5003f-faaa-40ac-939c-615c0da69deb/mouse-medium.jpg"
        @users[1].nickname.should == "test.user"
        @users[2].last_name.should == "service"
      end
            
      it 'should find a list of users through User model' do
        @users = User.search(:q => 'e')
        @users.size.should == 3
        @users[0].uuid.should == "7ee5003f-faaa-40ac-939c-615c0da69deb"
        @users[1].uuid.should == "2632c516-eb32-4cc8-879f-e8060577fc14"
        @users[2].uuid.should == "ad80daad-7a2f-4f5e-a62e-3189248486a9"
        @users[0].photo.should == "http://localhost:3000/uploads/user/photo/7ee5003f-faaa-40ac-939c-615c0da69deb/mouse-medium.jpg"
        @users[1].nickname.should == "test.user"
        @users[2].last_name.should == "service"
      end      
      
    end # context 'when search'
  end # context 'with remote user'
end
